@@ -2,7 +2,10 @@ from __future__ import division
 
 import pycparser.c_parser
 import pycparser.c_ast as c_ast
-import ply.yacc
+try:
+    import pycparser.ply.yacc as yacc
+except ImportError:
+    import ply.yacc as yacc
 
 
 class CParserBase(pycparser.c_parser.CParser):
@@ -15,6 +18,7 @@ class CParserBase(pycparser.c_parser.CParser):
         'expression',
         'identifier_list',
         'init_declarator_list',
+        'initializer_list',
         'parameter_type_list',
         'specifier_qualifier_list',
         'block_item_list',
@@ -35,14 +39,9 @@ class CParserBase(pycparser.c_parser.CParser):
         for rule in self.OPT_RULES:
             self._create_opt_rule(rule)
 
-        if hasattr(self, "p_translation_unit_or_empty"):
-            # v2.08 and later
-            self.ext_start_symbol = "translation_unit_or_empty"
-        else:
-            # v2.07 and earlier
-            self.ext_start_symbol = "translation_unit"
+        self.ext_start_symbol = "translation_unit_or_empty"
 
-        self.cparser = ply.yacc.yacc(
+        self.cparser = yacc.yacc(
             module=self,
             start=self.ext_start_symbol,
             debug=yacc_debug, write_tables=False)
@@ -124,7 +123,7 @@ class PreprocessorLine(c_ast.Node):
     def children(self):
         return ()
 
-    attr_names = ("contents")
+    attr_names = ("contents",)
 
 
 class TypeOfDeclaration(c_ast.Node):
@@ -155,6 +154,31 @@ class TypeOfExpression(c_ast.Node):
     attr_names = ()
 
 
+# These are the same as pycparser's, but it does *not* declare __slots__--
+# so we can poke in attributes at our leisure.
+class TypeDeclExt(c_ast.TypeDecl):
+    @staticmethod
+    def from_pycparser(td):
+        assert isinstance(td, c_ast.TypeDecl)
+        return TypeDeclExt(td.declname, td.quals, td.type, td.coord)
+
+
+class ArrayDeclExt(c_ast.TypeDecl):
+    @staticmethod
+    def from_pycparser(ad):
+        assert isinstance(ad, c_ast.ArrayDecl)
+        return ArrayDeclExt(ad.type, ad.dim, ad.dim_quals, ad.coord)
+
+
+def to_decl_ext(d):
+    if isinstance(d, c_ast.TypeDecl):
+        return TypeDeclExt.from_pycparser(d)
+    elif isinstance(d, c_ast.ArrayDecl):
+        return ArrayDeclExt.from_pycparser(d)
+    else:
+        raise TypeError("unexpected decl type: %s" % type(d).__name__)
+
+
 class FuncDeclExt(c_ast.Node):
     def __init__(self, args, type, attributes, asm, coord=None):
         self.args = args
@@ -196,6 +220,7 @@ class _AttributesMixin(object):
 
     def p_attribute_decl(self, p):
         """ attribute_decl : __ATTRIBUTE__ LPAREN LPAREN attribute_list RPAREN RPAREN
+                           | __ATTRIBUTE LPAREN LPAREN attribute_list RPAREN RPAREN
         """
         p[0] = p[4]
 
@@ -226,16 +251,56 @@ class _AttributesMixin(object):
         """ declarator  : direct_declarator attributes_opt
         """
         if p[2].exprs:
-            if not isinstance(p[1], c_ast.TypeDecl):
+            if isinstance(p[1], c_ast.ArrayDecl):
+                p[1].type = to_decl_ext(p[1].type)
+                p[1].type.attributes = p[2]
+            elif isinstance(p[1], c_ast.FuncDecl):
+                p[1].type = to_decl_ext(p[1].type)
+                p[1].type.attributes = p[2]
+            elif not isinstance(p[1], c_ast.TypeDecl):
                 raise NotImplementedError(
                         "cannot attach attributes to nodes of type '%s'"
                         % type(p[1]))
-
-            p[1].attributes = p[2]
+            else:
+                p[1] = to_decl_ext(p[1])
+                p[1].attributes = p[2]
 
         p[0] = p[1]
 
     # }}}
+    def p_declarator_2(self, p):
+        """declarator  : pointer direct_declarator attributes_opt
+                       | pointer attributes_opt direct_declarator
+        """
+        if hasattr(p[3], "exprs"):
+            attr_decl = p[3]
+            decl = p[2]
+        elif hasattr(p[2], "exprs"):
+            attr_decl = p[2]
+            decl = p[3]
+        else:
+            attr_decl = None
+            decl = p[2]
+
+        if not attr_decl.exprs:
+            attr_decl = None
+
+        if attr_decl:
+            if isinstance(decl, c_ast.ArrayDecl):
+                decl.type = to_decl_ext(decl.type)
+                decl.type.attributes = attr_decl
+            elif isinstance(decl, c_ast.FuncDecl):
+                decl.type = to_decl_ext(decl.type)
+                decl.type.attributes = attr_decl
+            elif not isinstance(p[2], c_ast.TypeDecl):
+                raise NotImplementedError(
+                        "cannot attach attributes to nodes of type '%s'"
+                        % type(p[1]))
+            else:
+                decl = to_decl_ext(decl)
+                decl.attributes = attr_decl
+
+        p[0] = self._type_modify_decl(decl, p[1])
 
     def p_function_specifier_attr(self, p):
         """ function_specifier  : attribute_decl
@@ -254,30 +319,30 @@ class _AsmMixin(object):
         p[0] = None
 
     def p_asm_opt_2(self, p):
-        """ asm_opt : asm
+        """ asm_opt : asm_no_semi
         """
         p[0] = p[1]
 
     def p_asm_1(self, p):
-        """ asm : asm_keyword LPAREN asm_argument_expression_list RPAREN
+        """ asm_no_semi : asm_keyword LPAREN asm_argument_expression_list RPAREN
         """
         p[0] = Asm(p[1], p[3], None, None, None, coord=self._coord(p.lineno(1)))
 
     def p_asm_2(self, p):
-        """ asm : asm_keyword LPAREN asm_argument_expression_list COLON \
+        """ asm_no_semi : asm_keyword LPAREN asm_argument_expression_list COLON \
                 asm_argument_expression_list RPAREN
         """
         p[0] = Asm(p[1], p[3], p[5], None, None, coord=self._coord(p.lineno(1)))
 
     def p_asm_3(self, p):
-        """ asm : asm_keyword LPAREN asm_argument_expression_list COLON \
+        """ asm_no_semi : asm_keyword LPAREN asm_argument_expression_list COLON \
                 asm_argument_expression_list COLON asm_argument_expression_list \
                 RPAREN
         """
         p[0] = Asm(p[1], p[3], p[5], p[7], None, coord=self._coord(p.lineno(1)))
 
     def p_asm_4(self, p):
-        """ asm : asm_keyword LPAREN asm_argument_expression_list COLON \
+        """ asm_no_semi : asm_keyword LPAREN asm_argument_expression_list COLON \
                 asm_argument_expression_list COLON asm_argument_expression_list \
                 COLON asm_argument_expression_list RPAREN
         """
@@ -305,10 +370,16 @@ class _AsmMixin(object):
         p[0] = p[1]
 
     def p_statement_gnu(self, p):
-        """ statement   : asm
+        """ statement   : asm_no_semi
+                        | asm_no_semi SEMI
         """
         p[0] = p[1]
 
+#    def p_asm_with_semi(self, p):
+#        """ asm : asm SEMI
+#        """
+#        p[0] = p[1]
+
 
 class _AsmAndAttributesMixin(_AsmMixin, _AttributesMixin):
     # {{{ /!\ names must match C parser to override
@@ -393,11 +464,35 @@ class GnuCParser(_AsmAndAttributesMixin, CParserBase):
         p[0] = c_ast.FuncCall(c_ast.ID(p[1], self._coord(p.lineno(1))),
                 TypeList([p[3], p[5]], self._coord(p.lineno(2))))
 
+    def p_gnu_statement_expression(self, p):
+        """ gnu_statement_expression : LPAREN compound_statement RPAREN
+        """
+        p[0] = p[2]
+
+    def p_gnu_primary_expression_6(self, p):
+        """ primary_expression : gnu_statement_expression """
+        p[0] = p[1]
+
+    def p_statement(self, p):
+        """ statement   : labeled_statement
+                        | expression_statement
+                        | compound_statement
+                        | selection_statement
+                        | iteration_statement
+                        | jump_statement
+                        | gnu_statement_expression
+        """
+        p[0] = p[1]
+
     def p_attribute_const(self, p):
         """ attribute : __CONST
         """
         p[0] = c_ast.ID(name="__const", coord=self._coord(p.lineno(1)))
 
+    def p_struct_declaration_list_1(self, p):
+        """ struct_declaration_list : empty """
+        p[0] = None
+
 # }}}
 
 
@@ -410,13 +505,15 @@ class OpenCLCParser(_AsmAndAttributesMixin, CParserBase):
                 "%s%d" % (base_name, count)
                 for base_name in [
                     'char', 'uchar', 'short', 'ushort', 'int', 'uint',
-                    'long', 'ulong', 'float', 'double']
+                    'long', 'ulong', 'float', 'double', 'half']
                 for count in [2, 3, 4, 8, 16]
                 ])
             | set([
                 "intptr_t", "uintptr_t",
                 "intmax_t", "uintmax_t",
-                "size_t", "ptrdiff_t"])
+                "size_t", "ptrdiff_t",
+                "uint", "ulong", "ushort", "uchar",
+                "half", "bool"])
             | set(["int%d_t" % bc for bc in INT_BIT_COUNTS])
             | set(["uint%d_t" % bc for bc in INT_BIT_COUNTS])
             | set(["int_least%d_t" % bc for bc in INT_BIT_COUNTS])
